
/********************************************
files.c

libmawk changes (C) 2009-2010, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991-94.  Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/
#ifndef _POSIX_SOURCE
/* needed on some systems to get strdup() */
#define _POSIX_SOURCE
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "mawk.h"
#include "files.h"
#include "memory.h"
#include "fin.h"

static FILE *tfopen(char *, char *);
static void efflush(mawk_state_t *, FILE *);
static void add_to_child_list(mawk_state_t *, int, int);
static struct child *remove_from_child_list(mawk_state_t *, int);
extern int isatty(int);

#ifdef	V7
#include  <sgtty.h>							/* defines FIOCLEX */
#endif


#ifndef	 NO_FCNTL_H

#include <fcntl.h>
#define	 CLOSE_ON_EXEC(fd)    fcntl(fd, F_SETFD, 1)

#else
#define	 CLOSE_ON_EXEC(fd) ioctl(fd, FIOCLEX, (PTR) 0)
#endif

/* mawk_find a file on file_list */
PTR mawk_file_find(mawk_state_t *MAWK, STRING *sval, int type)
{
	register FILE_NODE *p = MAWK->file_list;
	FILE_NODE *q = (FILE_NODE *) 0;
	char *name = sval->str;
	char *ostr;

	while (1) {
		if (!p) {
			/* open a new one */
			p = ZMALLOC(FILE_NODE);

			switch (p->type = type) {
			case F_TRUNC:
				ostr = "w";
				if (!(p->ptr = (PTR) tfopen(name, ostr)))
					goto out_failure;
				break;

			case F_APPEND:
				ostr = "a";
				if (!(p->ptr = (PTR) tfopen(name, ostr)))
					goto out_failure;
				break;

			case F_IN:
				if (!(p->ptr = (PTR) mawk_FINopen(MAWK, name, 0))) {
					zfree(MAWK, p, sizeof(FILE_NODE));
					return (PTR) 0;
				}
				break;

			case PIPE_OUT:
			case PIPE_IN:
				if (!(p->ptr = mawk_get_pipe(MAWK, name, type, &p->pid))) {
					if (type == PIPE_OUT) {
						goto out_failure;
					}
					else {
						zfree(MAWK, p, sizeof(FILE_NODE));
						return (PTR) 0;
					}
				}
				break;

#ifdef	DEBUG
			default:
				mawk_bozo(MAWK, "bad file type");
#endif
			}
			/* successful open */
			p->name = sval;
			sval->ref_cnt++;
			break;										/* while loop */
		}

		/* search is by name and type */
		if (((strcmp(name, p->name->str) == 0 && (p->type == type)) ||
																						/* no distinction between F_APPEND and F_TRUNC here */
																						(p->type >= F_APPEND && type >= F_APPEND)))
		{
			/* found */
			if (!q)										/*at front of list */
				return p->ptr;
			/* delete from list for move to front */
			q->link = p->link;
			break;										/* while loop */
		}

		q = p;
		p = p->link;
	}															/* end while loop */

	/* put p at the front of the list */
	p->link = MAWK->file_list;
	return (PTR) (MAWK->file_list = p)->ptr;

out_failure:
	mawk_errmsg(MAWK, errno, "cannot open \"%s\" for output", name);
	mawk_exitval(MAWK, 2, NULL);

}


/* Close a file and delete it's node from the file_list.
   Walk the whole list, in case a name has two nodes,
   e.g. < "/dev/tty" and > "/dev/tty"
*/
int mawk_file_close(mawk_state_t *MAWK, STRING *sval)
{
	FILE_NODE dummy;
	register FILE_NODE *p;
	FILE_NODE *q = &dummy;				/* trails p */
	FILE_NODE *hold;
	char *name = sval->str;
	int retval = -1;

	dummy.link = p = MAWK->file_list;
	while (p) {
		if (strcmp(name, p->name->str) == 0) {
			/* found */
			switch (p->type) {
			case F_TRUNC:
			case F_APPEND:
				fclose((FILE *) p->ptr);
				retval = 0;
				break;

			case PIPE_OUT:
				fclose((FILE *) p->ptr);
				retval = mawk_wait_for(MAWK, p->pid);
				break;

			case F_IN:
				mawk_FINclose(MAWK, (FIN *) p->ptr);
				retval = 0;
				break;

			case PIPE_IN:
				mawk_FINclose(MAWK, (FIN *) p->ptr);
				retval = mawk_wait_for(MAWK, p->pid);
				break;
			}

			free_STRING(p->name);
			hold = p;
			q->link = p = p->link;
			ZFREE(hold);
		}
		else {
			q = p;
			p = p->link;
		}
	}

	MAWK->file_list = dummy.link;
	return retval;
}

/*
mawk_find an output file with name == sval and fflush it
*/
int mawk_file_flush(mawk_state_t *MAWK, STRING *sval)
{
	int ret = -1;
	register FILE_NODE *p = MAWK->file_list;
	unsigned len = sval->len;
	char *str = sval->str;

	if (len == 0) {
		/* for consistency with gawk */
		mawk_flush_all_output(MAWK);
		return 0;
	}

	while (p) {
		if (IS_OUTPUT(p->type) && len == p->name->len && strcmp(str, p->name->str) == 0) {
			ret = 0;
			efflush(MAWK, (FILE *) p->ptr);
			/* it's possible for a command and a file to have the same
			   name -- so keep looking */
		}
		p = p->link;
	}
	return ret;
}

void mawk_flush_all_output(mawk_state_t * MAWK)
{
	FILE_NODE *p;

	for (p = MAWK->file_list; p; p = p->link)
		if (IS_OUTPUT(p->type))
			efflush(MAWK, (FILE *) p->ptr);
}

static void efflush(mawk_state_t *MAWK, FILE *fp)
{
	if (fflush(fp) < 0) {
		mawk_errmsg(MAWK, errno, "unexpected write error");
		mawk_exit(MAWK, 2);
	}
}


/* When we exit, we need to close and wait for all output pipes */

#if MAWK_NO_PIPE
void mawk_close_out_pipes(mawk_state_t * MAWK)
{
	if (!MAWK->do_exit)
		mawk_rt_error(MAWK, "pipes not supported");
}

PTR mawk_get_pipe(mawk_state_t *MAWK, char *name, int type, int *pid_ptr)
{
	if (!MAWK->do_exit)
		mawk_rt_error(MAWK, "pipes not supported");
	return NULL;
}

#else

/* work around for bug in AIX 4.1 -- If there are exactly 16 or 
   32 or 48 ..., open files then the last one doesn't get flushed on
   exit.  So the following is now a misnomer as we'll really close
   all output.
*/

void mawk_close_out_pipes(mawk_state_t * MAWK)
{
	register FILE_NODE *p = MAWK->file_list;

	while (p) {
		if (IS_OUTPUT(p->type)) {
			if (((p->ptr != stdout) && (p->ptr != stderr)) || (!MAWK->dont_close_std))
				fclose((FILE *) p->ptr);
			if (p->type == PIPE_OUT)
				mawk_wait_for(MAWK, p->pid);
		}
		p = p->link;
	}
}

PTR mawk_get_pipe(mawk_state_t *MAWK, char *name, int type, int *pid_ptr)
{
	int the_pipe[2], local_fd, remote_fd;

	if (pipe(the_pipe) == -1)
		return (PTR) 0;
	local_fd = the_pipe[type == PIPE_OUT];
	remote_fd = the_pipe[type == PIPE_IN];
	/* to keep output ordered correctly */
	fflush(stdout);
	fflush(stderr);

	switch (*pid_ptr = fork()) {
	case -1:
		close(local_fd);
		close(remote_fd);
		return (PTR) 0;

	case 0:
		close(local_fd);
		close(type == PIPE_IN);
		dup(remote_fd);
		close(remote_fd);
		execl(MAWK->shell, MAWK->shell, "-c", name, (char *) 0);
		mawk_errmsg(MAWK, errno, "failed to exec %s -c %s", MAWK->shell, name);
		fflush(stderr);
		_exit(128);

	default:
		close(remote_fd);
		/* we could deadlock if future child inherit the local fd ,
		   set close on exec flag */
		CLOSE_ON_EXEC(local_fd);
		break;
	}

	return type == PIPE_IN ? (PTR) mawk_FINdopen(MAWK, local_fd, 0) : (PTR) fdopen(local_fd, "w");
}
#endif /* MAWK_NO_PIPE */

/*------------ children ------------------*/

/* we need to wait for children at the end of output pipes to
   complete so we know any files they have created are complete */

static void add_to_child_list(mawk_state_t *MAWK, int pid, int exit_status)
{
	register struct child *p = ZMALLOC(struct child);

	p->pid = pid;
	p->exit_status = exit_status;
	p->link = MAWK->child_list;
	MAWK->child_list = p;
}

static struct child *remove_from_child_list(mawk_state_t *MAWK, int pid)
{
	struct child dummy;
	register struct child *p;
	struct child *q = &dummy;

	dummy.link = p = MAWK->child_list;
	while (p) {
		if (p->pid == pid) {
			q->link = p->link;
			break;
		}
		else {
			q = p;
			p = p->link;
		}
	}

	MAWK->child_list = dummy.link;
	return p;
	/* null return if not in the list */
}


/* wait for a specific child to complete and return its
   exit status

   If pid is zero, wait for any single child and
   put it on the dead children list
*/
int mawk_wait_for(mawk_state_t *MAWK, int pid)
{
	int exit_status;
	struct child *p;
	int id;

	if (pid == 0) {
		id = wait(&exit_status);
		add_to_child_list(MAWK, id, exit_status);
	}
	/* see if an earlier wait() caught our child */
	else if ((p = remove_from_child_list(MAWK, pid))) {
		exit_status = p->exit_status;
		ZFREE(p);
	}
	else {
		/* need to really wait */
		while ((id = wait(&exit_status)) != pid) {
			if (id == -1)							/* can't happen */
				mawk_bozo(MAWK, "mawk_wait_for");
			else {
				/* we got the exit status of another child
				   put it on the child list and try again */
				add_to_child_list(MAWK, id, exit_status);
			}
		}
	}

	if (exit_status & 0xff)
		exit_status = 128 + (exit_status & 0xff);
	else
		exit_status = (exit_status & 0xff00) >> 8;

	return exit_status;
}



void mawk_set_stderr(mawk_state_t * MAWK)
{																/* and stdout */
	FILE_NODE *p, *q;

	p = ZMALLOC(FILE_NODE);
	p->link = (FILE_NODE *) 0;
	p->type = F_TRUNC;
	p->name = mawk_new_STRING(MAWK, "/dev/stdout");
	p->ptr = (PTR) stdout;
	q = ZMALLOC(FILE_NODE);
	q->link = p;
	q->type = F_TRUNC;
	q->name = mawk_new_STRING(MAWK, "/dev/stderr");
	q->ptr = (PTR) stderr;
	MAWK->file_list = q;
}

/* fopen() but no buffering to ttys */
static FILE *tfopen(char *name, char *mode)
{
	FILE *retval = fopen(name, mode);

	if (retval) {
		if (isatty(fileno(retval)))
			setbuf(retval, (char *) 0);
	}
	return retval;
}

