
/********************************************
init.c

libmawk changes (C) 2009-2010, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

#include <stdlib.h>

#ifndef _POSIX_SOURCE
/* needed on some systems to get strdup() */
#define _POSIX_SOURCE
#endif

#include <string.h>
#include "mawk.h"
#include "code.h"
#include "memory.h"
#include "symtype.h"
#include "init.h"
#include "scan.h"
#include "bi_vars.h"
#include "field.h"
#include "zmalloc.h"

static int process_cmdline(mawk_state_t *, int, char **);
static void set_ARGV(mawk_state_t *, int, char **, int);
static void bad_option(mawk_state_t *, char *);
static void no_program(mawk_state_t *);

extern void mawk_print_version(void);
extern int mawk_is_cmdline_assign(mawk_state_t *, char *);

#ifndef	 SET_PROGNAME
#define	 SET_PROGNAME() \
   {char *p = strrchr(argv[0],'/') ;\
    MAWK->progname = p ? p+1 : argv[0] ; }
#endif

static const STRING null_str_ = { 0, 1, "" };	/* read-only */

static const mawk_escape_t escape_test_[ET_END + 1] =	/* read-only */
{
	{'n', '\n'},
	{'t', '\t'},
	{'f', '\f'},
	{'b', '\b'},
	{'r', '\r'},
	{'a', '\07'},
	{'v', '\013'},
	{'\\', '\\'},
	{'\"', '\"'},
	{0, 0}
};

mawk_state_t *mawk_initialize(int argc, char **argv)
{
	int mpow2[NUM_CELL_TYPES] = { 1, 2, 4, 8, 16, 32, 64, 128, 256, 512 };
	SEPARATOR rs_ = { SEP_CHAR, '\n' };
	CELL fs_ = { C_SPACE };

	mawk_state_t *MAWK;
	MAWK = calloc(sizeof(mawk_state_t), 1);

	MAWK->last_token_lineno = -1;

	MAWK->null_str = null_str_;
	MAWK->argi = NUM_ONE;
	MAWK->current_token = -1;
	MAWK->execution_start = 0;
	MAWK->ps.code_move_level = 0;
	MAWK->stack_base = MAWK->eval_stack;	/* these can move for deep recursion */
	MAWK->stack_danger = MAWK->eval_stack + DANGER;
	memcpy(MAWK->escape_test, escape_test_, sizeof(escape_test_));
	MAWK->interactive_flag = 0;
	MAWK->shell = strdup("/bin/sh");
	MAWK->max_field = MAX_SPLIT;
	MAWK->rs_shadow = rs_;
	MAWK->fs_shadow = fs_;
	memcpy(MAWK->scan_code, mawk_scan_code, sizeof(mawk_scan_code));

	/* this can be moved and enlarged  by -W sprintf=num  */
	MAWK->sprintf_buff = string_buff;
	MAWK->sprintf_limit = string_buff + sizeof(MAWK->tempbuff);
	MAWK->mpow2 = malloc(sizeof(mpow2));
	memcpy(MAWK->mpow2, mpow2, sizeof(mpow2));
	MAWK->fbank[0] = MAWK->field;

	SET_PROGNAME();

	mawk_bi_vars_init(MAWK);			/* load the builtin variables */
	mawk_bi_funct_init(MAWK);			/* load the builtin functions */
	mawk_kw_init(MAWK);						/* load the keywords */
	mawk_field_init(MAWK);

	if (!process_cmdline(MAWK, argc, argv))
		return NULL;

	mawk_code_init(MAWK);
	mawk_fpe_init();
	mawk_set_stderr(MAWK);

	return MAWK;
}

void mawk_hash_clear(mawk_state_t *MAWK);

void mawk_uninitialize(mawk_state_t * m)
{
	/* free global variables */
	mawk_hash_clear(m);

	if (m->Argv != NULL)
		free(m->Argv->ptr);

	if (m->shell != NULL)
		free(m->shell);
	if (m->mpow2 != NULL)
		free(m->mpow2);

	mawk_bi_vars_uninit(m);


	if (m->mawk_parser_stack != NULL)
		free(m->mawk_parser_stack);


	if (m->ps.buffer != NULL)
		zfree(m, m->ps.buffer, BUFFSZ + 1);

	if (m->scripts_loaded != NULL)
		mawk_array_clear(m, m->scripts_loaded);

	mawk_free_all(m);
	free(m);
}

static void bad_option(mawk_state_t *MAWK, char *s)
{
	mawk_errmsg(MAWK, 0, "not an option: %s", s);
	mawk_exit(MAWK, 2);
}

static void no_program(mawk_state_t * MAWK)
{
	mawk_exit(MAWK, 0);
}

void mawk_append_input_file(mawk_state_t * MAWK, const char *fn, int bytecode)
{
	/* first file goes in pfile_name ; any more go on a list */
	if (MAWK->ps.pfile_name) {
		if (MAWK->pfile_list_tail == NULL) {
			MAWK->pfile_list_tail = ZMALLOC(PFILE);
			MAWK->pfile_list_tail->fname = fn;
			MAWK->pfile_list_tail->bytecode = bytecode;
			MAWK->pfile_list = MAWK->pfile_list_tail;
		}
		else {
			MAWK->pfile_list_tail->link = ZMALLOC(PFILE);
			MAWK->pfile_list_tail = MAWK->pfile_list_tail->link;
			MAWK->pfile_list_tail->fname = NULL;
			MAWK->pfile_list_tail->bytecode = 0;
		}
		MAWK->pfile_list_tail->link = NULL;
	}
	else {
		MAWK->ps.pfile_name = fn;
		MAWK->ps.pfile_bytecode = bytecode;
	}
}

static int process_cmdline(mawk_state_t *MAWK, int argc, char **argv)
{
	int i, nextarg;
	char *optarg;
	MAWK->pfile_list = NULL;
	MAWK->pfile_list_tail = NULL;

	for (i = 1; i < argc && argv[i][0] == '-'; i = nextarg) {
		if (argv[i][1] == 0) {			/* -  alone */
			if (!MAWK->ps.pfile_name) {
				no_program(MAWK);
				return 0;
			}
			break;										/* the for loop */
		}
		/* safe to look at argv[i][2] */

		if (argv[i][2] == 0) {
			if (i == argc - 1 && argv[i][1] != '-') {
				if (strchr("WFvf", argv[i][1])) {
					mawk_errmsg(MAWK, 0, "option %s lacks argument", argv[i]);
					mawk_exit_(MAWK, 2);
					return 0;
				}
				bad_option(MAWK, argv[i]);
			}

			optarg = argv[i + 1];
			nextarg = i + 2;
		}
		else {											/* argument glued to option */

			optarg = &argv[i][2];
			nextarg = i + 1;
		}

		switch (argv[i][1]) {
		case 'W':

			if (optarg[0] >= 'a' && optarg[0] <= 'z')
				optarg[0] += 'A' - 'a';
			if (optarg[0] == 'V')
				mawk_print_version();
			else if (optarg[0] == 'C')
					MAWK->dump_code_flag = 2;
			else if (optarg[0] == 'D') {
				if ((optarg[1] == 'U') || (optarg[1] == 'u'))
					MAWK->dump_code_flag = 1;
				else
					MAWK->debug_symbols = 1;
			}
			else if (optarg[0] == 'S') {
				char *p = strchr(optarg, '=');
				int x = p ? atoi(p + 1) : 0;

				if (x > SPRINTF_SZ) {
					MAWK->sprintf_buff = (char *) zmalloc(MAWK, x);
					MAWK->sprintf_limit = MAWK->sprintf_buff + x;
				}
			}
			else if (optarg[0] == 'P') {
				MAWK->posix_space_flag = 1;
			}
			else if (optarg[0] == 'E') {
				if (MAWK->ps.pfile_name) {
					mawk_errmsg(MAWK, 0, "-W exec is incompatible with -f");
					mawk_exit_(MAWK, 2);
					return 0;
				}
				else if (nextarg == argc)
					no_program(MAWK);

				MAWK->ps.pfile_name = argv[nextarg];
				MAWK->ps.pfile_bytecode = 0;
				i = nextarg + 1;
				goto no_more_opts;
			}
			else if (optarg[0] == 'I') {
				MAWK->interactive_flag = 1;
				setbuf(stdout, (char *) 0);
			}
			else if (strncmp(optarg, "Maxmem=", 7) == 0) {
				char *end;
				MAWK->mm_max = strtol(optarg+7, &end, 10);
				switch(*end) {
					case '\0':
						break;
					case 'k':
					case 'K':
						MAWK->mm_max *= 1024;
						break;
					case 'm':
					case 'M':
						MAWK->mm_max *= 1024 * 1024;
						break;
					default:
						MAWK->mm_max = 0;
						mawk_errmsg(MAWK, 0, "invalid memory size for -Wmaxmem (must be integer with optional K or M suffix): '%s'", optarg+7);
				}
			}
			else
				mawk_errmsg(MAWK, 0, "vacuous option: -W %s", optarg);


			break;

		case 'v':
			if (!mawk_is_cmdline_assign(MAWK, optarg)) {
				mawk_errmsg(MAWK, 0, "improper assignment: -v %s", optarg);
				mawk_exit_(MAWK, 2);
				return 0;
			}
			break;

		case 'F':

			mawk_rm_escape(MAWK, optarg);	/* recognize escape sequences */
			cell_destroy(FS);
			FS->type = C_STRING;
			FS->ptr = (PTR) mawk_new_STRING(MAWK, optarg);
			mawk_cast_for_split(MAWK, mawk_cellcpy(MAWK, &MAWK->fs_shadow, FS));
			break;

		case '-':
			if (argv[i][2] != 0)
				bad_option(MAWK, argv[i]);
			i++;
			goto no_more_opts;

		case 'f':
			mawk_append_input_file(MAWK, optarg, 0);
			break;
		case 'b':
			mawk_append_input_file(MAWK, optarg, 1);
			break;
		default:
			bad_option(MAWK, argv[i]);
		}
	}

no_more_opts:
	if (MAWK->ps.pfile_name) { /* program from -f or -b */
		set_ARGV(MAWK, argc, argv, i);
		mawk_scan_init(MAWK, (char *) 0);
		if (MAWK->do_exit)
			return 0;
	}
	else {												/* program given on command line (no -f or -b) */

		if (i == argc) {
			no_program(MAWK);
			return 0;
		}
		set_ARGV(MAWK, argc, argv, i + 1);
		mawk_scan_init(MAWK, argv[i]);
		if (MAWK->do_exit)
			return 0;

/* #endif  */
	}
	return 1;
}


static void set_ARGV(mawk_state_t *MAWK, int argc, char **argv, int i)
{
	/* argv[i] = ARGV[i] */
	SYMTAB *st_p;
	CELL argi;
	register CELL *cp;

	st_p = mawk_insert(MAWK, "ARGV");
	st_p->type = ST_ARRAY;
	MAWK->Argv = st_p->stval.array = new_ARRAY();
	argi.type = C_NUM;
	argi.dval = NUM_ZERO;
	cp = mawk_array_find(MAWK, st_p->stval.array, &argi, CREATE);
	if (cp == NULL)
		return;
	cp->type = C_STRING;
	cp->ptr = (PTR) mawk_new_STRING(MAWK, MAWK->progname);

	/* ARGV[0] is set, do the rest
	   The type of ARGV[1] ... should be C_MBSTRN
	   because the user might enter numbers from the command line */

	for (argi.dval = NUM_ONE; i < argc; i++, argi.dval += NUM_ONE) {
		cp = mawk_array_find(MAWK, st_p->stval.array, &argi, CREATE);
		cp->type = C_MBSTRN;
		cp->ptr = (PTR) mawk_new_STRING(MAWK, argv[i]);
	}
	ARGC->type = C_NUM;
	ARGC->dval = argi.dval;
}


/*----- ENVIRON ----------*/

void mawk_load_environ(mawk_state_t *MAWK, ARRAY ENV)
{
	CELL c;
	extern char **environ;
	register char **p = environ;	/* walks environ */
	char *s;											/* looks for the '=' */
	CELL *cp;											/* pts at ENV[&c] */

	c.type = C_STRING;

	while (*p) {
		if ((s = strchr(*p, '='))) {	/* shouldn't fail */
			int len = s - *p;
			c.ptr = (PTR) mawk_new_STRING0(MAWK, len);
			memcpy(string(&c)->str, *p, len);
			s++;

			cp = mawk_array_find(MAWK, ENV, &c, CREATE);
			cp->type = C_MBSTRN;
			cp->ptr = (PTR) mawk_new_STRING(MAWK, s);

			free_STRING(string(&c));
		}
		p++;
	}
}
