
/********************************************
rexp.c

libmawk changes (C) 2009-2010, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

/*  op precedence  parser for regular expressions  */

#include  "rexp.h"


/*  DATA   */
int REerrno;
char *REerrlist[] = { (char *) 0,
	/* 1  */ "missing '('",
	/* 2  */ "missing ')'",
	/* 3  */ "bad class -- [], [^] or [",
	/* 4  */ "missing operand",
	/* 5  */ "resource exhaustion -- regular expression too large",
	/* 6  */ "syntax error ^* or ^+"
};

/* E5 is very unlikely to occur */


/* This table drives the operator precedence parser */
static short table[8][8] = {

/*        0   |   CAT   *   +   ?   (   )   */
/* 0 */ { 0,  L,   L,   L,  L,  L,  L,  E1},
/* | */ { G,  G,   L,   L,  L,  L,  L,  G},
/* CAT*/{ G,  G,   G,   L,  L,  L,  L,  G},
/* * */ { G,  G,   G,   G,  G,  G,  E7, G},
/* + */ { G,  G,   G,   G,  G,  G,  E7, G},
/* ? */ { G,  G,   G,   G,  G,  G,  E7, G},
/* ( */ { E2, L,   L,   L,  L,  L,  L,  EQ},
/* ) */ { G,  G,   G,   G,  G,  G,  E7, G}
};


#define	 STACKSZ   64


static jmp_buf err_buf;					/*  used to trap on error */

void mawk_RE_error_trap(int x)
{
	REerrno = x;
	longjmp(err_buf, 1);
}


PTR mawk_REcompile(char *re)
{
	MACHINE m_stack[STACKSZ];
	struct op {
		int token;
		int prec;
	} op_stack[STACKSZ];
	register MACHINE *m_ptr;
	register struct op *op_ptr;
	register int t;

	/* do this first because it also checks if we have a
	   run time stack */
	mawk_RE_lex_init(re);

	if (*re == 0) {
		STATE *p = (STATE *) mawk_RE_malloc(sizeof(STATE));
		p->type = M_ACCEPT;
		return (PTR) p;
	}

	if (setjmp(err_buf))
		return (PTR) 0;
	/* we used to try to recover memory left on machine stack ;
	   but now m_ptr is in a register so it won't be right unless
	   we force it out of a register which isn't worth the trouble */

	/* initialize the stacks  */
	m_ptr = m_stack - 1;
	op_ptr = op_stack;
	op_ptr->token = 0;

	t = mawk_RE_lex(m_stack);

	while (1) {
		switch (t) {
		case T_STR:
		case T_ANY:
		case T_U:
		case T_START:
		case T_END:
		case T_CLASS:
			m_ptr++;
			break;

		case 0:										/*  end of reg expr   */
			if (op_ptr->token == 0) {
				/*  done   */
				if (m_ptr == m_stack)
					return (PTR) m_ptr->start;
				else {
					/* machines still on the stack  */
					mawk_RE_panic("values still on machine stack");
				}
			}

			/*  otherwise  fall  thru to default
			   which is operator case  */

		default:

			if ((op_ptr->prec = table[op_ptr->token][t]) == G) {
				do {										/* op_pop   */

					if (op_ptr->token <= T_CAT)	/*binary op */
						m_ptr--;
					/* if not enough values on machine stack
					   then we have a missing operand */
					if (m_ptr < m_stack)
						mawk_RE_error_trap(-E4);

					switch (op_ptr->token) {
					case T_CAT:
						mawk_RE_cat(m_ptr, m_ptr + 1);
						break;

					case T_OR:
						mawk_RE_or(m_ptr, m_ptr + 1);
						break;

					case T_STAR:
						mawk_RE_close(m_ptr);
						break;

					case T_PLUS:
						mawk_RE_poscl(m_ptr);
						break;

					case T_Q:
						mawk_RE_01(m_ptr);
						break;

					default:
						/*nothing on ( or ) */
						break;
					}

					op_ptr--;
				}
				while (op_ptr->prec != L);

				continue;								/* back thru switch at top */
			}

			if (op_ptr->prec < 0) {
				if (op_ptr->prec == E7)
					mawk_RE_panic("parser returns E7");
				else
					mawk_RE_error_trap(-op_ptr->prec);
			}

			if (++op_ptr == op_stack + STACKSZ) {
				/* stack overflow */
				mawk_RE_error_trap(-E5);
			}

			op_ptr->token = t;
		}														/* end of switch */

		if (m_ptr == m_stack + (STACKSZ - 1)) {
			/*overflow */
			mawk_RE_error_trap(-E5);
		}

		t = mawk_RE_lex(m_ptr + 1);
	}
}


/* getting here means a logic flaw or unforeseen case */
void mawk_RE_panic(char *s)
{
	fprintf(stderr, "mawk_REcompile() - panic:  %s\n", s);
	exit(100);
}
